#include <time.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdarg.h>

#include "../util.h"
#include "../config.h"
#include "../crc16.h"
#include "../util.h"

#include "AM01_packet.h"
#include "AM01_protocol.h"

#define COORD(X)  SWAP_UINT32(X) / 1800000.0f;

void print_gps_info(connection * conn, information_package package) {
    gps_information info;
    memcpy(&info, package.information, sizeof(gps_information));
    float lat = COORD(info.lattitude);
    float lon = COORD(info.longitude) ;
    log_line(conn, "got gps info, lat/long: %f %f\n", lat, lon);
    gpsprintf(conn,
              "20%02u-%02u-%02uT%02u:%02u:%02uZ,%f,%f,%u,%u,%u,%u\n",
              package.date_time.year,
              package.date_time.month,
              package.date_time.day,
              package.date_time.hour,
              package.date_time.minute,
              package.date_time.second,
              lat,
              lon,
              info.speed,
              info.num_satelites,
              info.status,
              info.course);
}

void print_status_info(connection * conn, status_information status) {
    bool fortified = status.device_info & 1;
    bool acc = status.device_info & 2;
    bool charged = status.device_info & 4;
    uint8_t alm = status.device_info >> 3 & 7;
    bool located = status.device_info & 0x40;
    bool power = status.device_info & 0x80;
    log_line(conn, "got status package, acc: %u  charged: %u alarm:%u located: %u power: %u voltage: %u\n", fortified, acc, charged, alm, located, power,        status.voltage);
}

void process_alarm(connection * conn, information_package package) {
    gps_lbs_status_information info;
    uint8_t lbs_size = *(package.information + sizeof(gps_information));
    memcpy(&info.gps_info, package.information, sizeof(gps_information));
    memcpy(&info.lbs_info, package.information + 1 + sizeof(gps_information), min(lbs_size, sizeof(lbs_information)));
    memcpy(&info.status_info, package.information + 1 + sizeof(gps_information) + lbs_size, sizeof(status_information));
    float lat = COORD(info.gps_info.lattitude);
    float lon = COORD(info.gps_info.longitude);
    bool speeding = (info.status_info.alarm == 0 );
    bool escape = (info.status_info.alarm == 2) ;
    const char * typemsg = speeding ? "speeding" : "unknown";

    if (escape) {
        typemsg = "tracker removed";
    }

    log_line(conn, "got alarm, info %x %x %x %x\n", info.status_info.device_info, info.status_info.voltage, info.status_info.gsm_strength, info.status_info.alarm);
    log_line(conn, "got alarm, lat/long: %f %f\n", lat, lon);
    eventprintf(conn,
                "20%02u-%02u-%02uT%02u:%02u:%02uZ,%f,%f,%u,%s\n",
                package.date_time.year,
                package.date_time.month,
                package.date_time.day,
                package.date_time.hour,
                package.date_time.minute,
                package.date_time.second,
                lat,
                lon,
                info.gps_info.speed,
                typemsg);
    print_status_info(conn, info.status_info);
}

void process_information_package(connection * conn, data_packet packet) {
    information_package package;
    int dataLength = packet.header.length - 5;
    memcpy(&package, packet.data, min(dataLength, sizeof(date_time_info)));
    memcpy(package.information, packet.data + sizeof(date_time_info), dataLength - sizeof(date_time_info));
    log_line(conn, "got information package %x data starting with %x %x\n", packet.header.protocol_number, package.information[0], package.information[1]);

    switch (packet.header.protocol_number) {
        case 0x22:
        case 0x10:	//gps information
            print_gps_info(conn, package);
            break;

        case 0x12:	//gps/lbs merged information	rm test 2>/dev/null
            log_line(conn, "got gps/lbs merged info \n");
            print_gps_info(conn, package);
            break;

        case 0x16:	//gps/lbs status merged information ( alarm packet )
        case 0x26: //alarm packet specific for this device
            process_alarm(conn, package);
            break;

        default:
            log_line(conn, "weird information package, protocol number: %u\n", packet.header.protocol_number);
    }
}

void process_v2(connection * conn) {
    uint8_t response[BUF_SIZE];
    size_t length = data_length(PACKET(conn));
    memset(response, 0, BUF_SIZE);
    memcpy(response, PACKET(conn).data + 5, length - 5);
    log_current_packet_ascii(conn);

    //this is a long packet. for some reason every 251 bytes they've added a header. we need to remove it from our data to parse it.
    for (int i = 1; i < length; i++) {
        if (i % 251 == 0) {
            memmove(&response[i], &response[i + 6], length - i - 6);
            length -= 6;
        }
    }

    if (PACKET(conn).v2_header.protocol_number == 0x21) {
        log_line(conn, "command response of length %u : %s\n", data_length(PACKET(conn)), response);
        log_command_response(conn, response);

        if (memcmp(response, "Battery:", 8) == 0) {
            statusprintf(conn, "%s\n", response);
        }
    }
}

void process_v1(connection * conn) {
    int dataLength =  data_length(PACKET(conn));
    bool sendDefaults = false;

    switch (PACKET(conn).header.protocol_number) {
        case 0x01:	//login information
            sendDefaults = (fileModifiedAgo(conn->log_outfile) > 300);
            convert_imei(PACKET(conn).data, conn->imei);
            send_data_packet(conn, create_response(PACKET(conn).header.protocol_number, PACKET(conn).footer.serial_number));
            init_imei(conn);
            break;

        case 0x10:	//gps information
        case 0x11:	//lbs information
        case 0x22:
        case 0x12:	//gps/lbs merged information
            process_information_package(conn, PACKET(conn));
            break;

        case 0x13:	//status information ( keepalive packet )
            print_status_info(conn, *((status_information *)PACKET(conn).data));
            send_data_packet(conn, create_response(PACKET(conn).header.protocol_number, PACKET(conn).footer.serial_number));
            break;

        case 0x16:	//gps/lbs status merged information ( alarm packet )
        case 0x26: //alarm packet specific for this device
            process_information_package(conn, PACKET(conn));
            break;

        default:
            break;
    }
}

void process_current_packet(connection * conn) {
    //test if crc and end bytes valid. not for a v2 packet because the checksum for that is weird.
    if (PACKET(conn).footer.stop_bit[0] != 0xD ||
            PACKET(conn).footer.stop_bit[1] != 0xA || (
                !is_v2(PACKET(conn)) && PACKET(conn).footer.crc != crc16(PACKET(conn)))) {
        log_line(conn, "invalid footer bytes or checksum.\n");
    }

    switch (PACKET(conn).header.start_bit[0]) {
        case 0x79:
            process_v2(conn);
            break;

        case 0x78:
            process_v1(conn);
            break;

        default:
            break;
    }
}

void process_command_file(connection * conn) {
    char buffer[BUF_SIZE];
    FILE * fp = fopen(conn->command_infile, "r");

    //if there's no commands file, well there is nothing to do
    if (fp <= 0) {
        return;
    }

    memset(buffer, 0, BUF_SIZE);

    //read the file line by line and send our commands
    while (fgets(buffer, BUF_SIZE - 1, fp)) {
        if (strlen(buffer) > 2) {
            send_command(conn, buffer);
        }
    }

    //clear our commands file
    freopen(NULL, "w", fp);
    fclose(fp);
}


bool validate_startbits(data_packet p) {
    const uint8_t v1_start[] = {0x78, 0x78};
    const uint8_t v2_start[] = {0x79, 0x79};
    return memcmp(p.header.start_bit, v1_start, 2) == 0 || memcmp(p.header.start_bit, v2_start, 2) == 0 ;
}

void process(connection * conn) {
    conn->current_packet_valid = false;

    //nothing to do? process external commands.
    if ( conn->send_count == 0 && conn->read_count == 0 && conn->can_log) {
        process_command_file(conn);
    }

    //if we've at least got a header
    if (conn->read_count > sizeof(data_packet_header)) {
        memcpy(&PACKET(conn).header, conn->recv_buffer, min(conn->read_count, sizeof(data_packet_header)));
        memcpy(&PACKET(conn).v2_header, conn->recv_buffer, min(conn->read_count, sizeof(data_packet_v2_header)));

        //with valid start bits
        if (!validate_startbits(PACKET(conn))) {
            log_line(conn, "invalid header starting with byte: %u\n", PACKET(conn).header.start_bit[0]);
            memmove(conn->recv_buffer, conn->recv_buffer + 1, conn->read_count - 1);
            conn->read_count -= 1;

        } else {
            size_t dataLength = data_length(PACKET(conn));
            size_t header_size = (is_v2(PACKET(conn)) ? sizeof(data_packet_v2_header) : sizeof(data_packet_header));
            size_t totalSize = header_size +  dataLength +  sizeof(data_packet_footer);

            //and maybe enough data to reconstruct our entire packet
            if (conn->read_count >= totalSize) {
                memcpy(PACKET(conn).data, conn->recv_buffer + header_size, dataLength);
                memcpy(&PACKET(conn).footer, conn->recv_buffer + header_size + dataLength, sizeof(data_packet_footer));
                log_line(conn, "packet recieved.\n");
                log_current_packet(conn);
                process_current_packet(conn);
            }

            memmove(conn->recv_buffer, conn->recv_buffer + totalSize, conn->read_count - totalSize);
            conn->read_count -= totalSize;
        }

    } else {
        //if we're idle, and 5 minutes have passed we should get status from our device
        if (conn->read_count == 0 && ( time(0) - conn->since_last_status ) > 300) {
            send_command(conn, "STATUS#");
            conn->since_last_status = time(0);
        }
    }
}
