#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <err.h>
#include <sys/select.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <signal.h>

#include "sock_server.h"
#include "connection.h"
#include "device/AM01_protocol.h"



//sets the non-blocking flag for an IO descriptor
void set_nonblock(int fd) {
    //retrieve all the flags for this file descriptor
    int fl = fcntl(fd, F_GETFL, 0);

    if (fl < 0) {
        fprintf(stderr, "Failed to get flags for file descriptor %d: %s\n", fd, strerror(errno));
        exit(1);
    }

    //add the non-blocking flag to this file descriptor's flags
    if (fcntl(fd, F_SETFL, fl | O_NONBLOCK) < 0) {
        fprintf(stderr, "Failed to set flags for file descriptor %d: %s\n", fd, strerror(errno));
        exit(1);
    }

    int optval = 1;
    socklen_t optlen = sizeof(optval);
    optlen = sizeof(optval);

    if (setsockopt(fd, SOL_SOCKET, SO_KEEPALIVE, &optval, optlen) < 0) {
        fprintf(stderr, "Failed to set keepalive\n");
    }

    fprintf(stdout, "SO_KEEPALIVE set on socket\n");
}

int create_server_sock(char * addr, int port) {
    int on = 1;
    static struct sockaddr_in client_addr;
    int sock = socket(AF_INET, SOCK_STREAM, 0);

    if (sock < 0) {
        fprintf(stderr, "Failed to create server socket.\n");
        return -1;
    }

    memset(&client_addr, '\0', sizeof(client_addr));
    client_addr.sin_family = AF_INET;
    client_addr.sin_addr.s_addr = inet_addr(addr);
    client_addr.sin_port = htons(port);

    if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, 4) < 0) {
        fprintf(stderr, "Failed to set socket address %s:%d\n", addr, port);
        return -1;
    }

    if (bind(sock, (struct sockaddr *) &client_addr, sizeof(client_addr)) < 0) {
        fprintf(stderr, "Failed to bind server socket on %s:%d\n", addr, port);
        return -1;
    }

    if (listen(sock, 5) < 0) {
        fprintf(stderr, "Failed to listen on %s:%d\n", addr, port);
        return -1;
    }

    fprintf(stdout, "Listening on %s port %d\n", addr, port);
    return sock;
}

int wait_for_client(int s) {
    struct sockaddr_in peer;
    socklen_t len = sizeof(struct sockaddr);
    fprintf(stdout, "Accepting connections with file descriptor: %d\n", s);
    int newsock = accept(s, (struct sockaddr *) &peer, &len);

    if (newsock < 0 && errno != EINTR) {
        fprintf(stdout, "Failed to accept connection with file descriptor %d: %s\n", s, strerror(errno));
        return -1;
    }

    struct hostent * hostinfo = gethostbyaddr((char *) &peer.sin_addr.s_addr, len, AF_INET);

    fprintf(stdout, "Incoming connection accepted from %s[%s]\n", !hostinfo ? "" : hostinfo->h_name, inet_ntoa(peer.sin_addr));

    set_nonblock(newsock);

    return (newsock);
}



void process_thread(connection conn) {
    struct timeval to = { 0, 100 };
    time_t since_packet = time(0);

    for (;;) {
        //if we've read data from our cmmand we need to send it
        if (conn.send_count) {
            int sent = write(conn.socket, conn.send_buffer, conn.send_count);

            //if an error happens during sending - the client probably gave up
            if (sent < 0 && errno != EWOULDBLOCK) {
                fprintf(stdout, "client disconnected: %s\n", strerror(errno));
                exit(1);
            }

            //reduce our buffer by the amount of data sent
            if (sent != conn.send_count) {
                memmove(conn.send_buffer, conn.send_buffer + sent, conn.send_count - sent);
            }

            conn.send_count -= sent;
        }

        //use select to wait until anything happens to our source file.
        fd_set R;
        FD_ZERO(&R);
        FD_SET(conn.socket, &R);
        int selectValue = select(conn.socket + 1, &R, 0, 0, &to);

        if (selectValue > 0) {
            int rdCount = read(conn.socket, conn.recv_buffer + conn.read_count, BUF_SIZE - conn.read_count);	//read potential input from the client - and discard it.

            //if our program closed - end the session
            if (rdCount < 0 && errno != EWOULDBLOCK) {
                fprintf(stdout, "socket closed : %s\n", strerror(errno));
                exit(1);
            }

            //add the amount of data to the buffer that we have to send
            if (rdCount > 0) {
                conn.read_count += rdCount;
                since_packet = time(0);
            }
        }

        process(&conn);

        //if our client disconnected for any reason
        if (selectValue < 0 && errno != EINTR) {
            fprintf(stdout, "client disconnected: %s\n", strerror(errno));
            close(conn.socket);
            close_connection(&conn);
            exit(0);
        }

        //test for disconnected client every 5 minutes then end the process
        if (time(0) - since_packet > 310) {
            fprintf(stdout, "client timed out\n");
            close(conn.socket);
            close_connection(&conn);
            exit(0);
        }

        usleep(GRACE_TIME);
    }
}

void run_server() {
    int client = -1;
    int server_socket ;

    while (server_socket <= 0) {
        server_socket = create_server_sock(LISTEN_ON, LISTEN_PORT);

        if (server_socket <= 0) {
            sleep(10000);
        }
    }

    //will not wait for child processes
    signal(SIGCHLD, SIG_IGN);

    for (;;) {
        if ((client = wait_for_client(server_socket)) < 0) {
            continue;
        }

        if (0 == fork()) {
            close(server_socket);
            connection conn = new_connection(client);
            process_thread(conn);
            abort();
        }

        close(client);
        client = -1;
    }
}
