#include <time.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <stdarg.h>

#include "../util.h"
#include "../config.h"
#include "../crc16.h"
#include "../string.h"
#include "MT200x_protocol.h"

void send_string(connection * conn, char * str) {
    strcpy(conn->send_buffer + conn->send_count, str);
    conn->send_count += strlen(str);
}

void send_command( connection * conn, char * cmd) {
    size_t start = conn->send_count;
    cmd = strip_whitespace(cmd);
    send_string(conn, "$GPRS,");
    send_string(conn, conn->imei + (*conn->imei == '0' ? 1 : 0));
    send_string(conn, ";");
    send_string(conn, cmd);
    send_string(conn, ";!");
    log_line(conn, "sent command: ");

    for (int i = start; i < conn->send_count; i++) {
        logprintf(conn, "%c", ((unsigned char *)conn->send_buffer)[i]);
    }

    logprintf(conn, "\n");
}

void process_command_file(connection * conn) {
    char buffer[BUF_SIZE];
    FILE * fp = fopen(conn->command_infile, "r");

    //if there's no commands file, well there is nothing to do
    if (fp <= 0) {
        return;
    }

    memset(buffer, 0, BUF_SIZE);

    //read the file line by line and send our commands
    while (fgets(buffer, BUF_SIZE - 1, fp)) {
        if (strlen(buffer) > 2) {
            send_command(conn, buffer);
        }
    }

    //clear our commands file
    freopen(NULL, "w", fp);
    fclose(fp);
}


float parseLat(char * str) {
    return parse_int(str, 2) + parse_float(str + 2) / 60.0f;
}

float parseLong(char * str) {
    return parse_int(str, 3) + parse_float(str + 3) / 60.0f;
}

void format_imei(char * imei) {
    for (; strlen(imei) < 16;) {
        memmove(imei + 1, imei, strlen(imei));
        imei[0] = '0';
    }
}

void process_message(connection * conn, char * string, size_t length) {
    unsigned char bufstr[BUF_SIZE];
    unsigned char * data_buffers[40];
    unsigned char imei[64] = {0};
    string = strip_whitespace(string);
    memset(imei, 0, 64);
    memset(bufstr, 0, sizeof(bufstr));
    memset(data_buffers, 0, sizeof(data_buffers));
    memcpy(bufstr, conn->recv_buffer, min(length, BUF_SIZE - 1));
    log_line(conn, "message recieved: %s\n", bufstr);
    size_t str_count = split_to(',', bufstr, BUF_SIZE, data_buffers, 40);

    if (strstr(data_buffers[0], "CMV001") || strstr(data_buffers[0], ";")) {
        memcpy(bufstr, conn->recv_buffer, min(length, BUF_SIZE - 1));
        log_command_response(conn, bufstr);
        return;
    }

    log_line(conn, "  parsed message: ");

    for (size_t i = 0; i < str_count; i++) {
        logprintf(conn, "  [%u]: %s ", i, data_buffers[i]);
    }

    logprintf(conn, "\n");

    if ( str_count < 35) {
        log_line(conn, "  invalid response length.\n");
        return;
    }

    memcpy(imei, data_buffers[1], min(strlen(data_buffers[1]), 63));
    format_imei(imei);

    if (strlen(conn->imei) <= 1 || strcmp(conn->imei, imei) != 0) {
        memcpy(conn->imei, imei, strlen(imei) + 1);
        init_imei(conn);
    }

    if ( strlen(data_buffers[7]) < 5 || strlen(data_buffers[9]) < 6 ) {
        log_line(conn, "  invalid lat or long coordinates.\n");
        return;
    }

    float lat = parseLat(data_buffers[7]);
    float lon = parseLong(data_buffers[9]);
    unsigned char * dtstr = data_buffers[4];
    unsigned char * timestr = data_buffers[5];
    float spd = parse_float(data_buffers[15]) * 1.852f;

    if ( strlen(dtstr) < 6 || strlen(timestr) < 6) {
        log_line(conn, "  invalid date or time.\n");
        return;
    }

    statusprintf(conn, "Battery: %s; GPS satelites: %s; Last status: %s; Last speed: %f\n",
                 data_buffers[33],
                 data_buffers[12],
                 data_buffers[34],
                 spd);

    if (strcmp(data_buffers[34], "Timer") == 0) {
        gpsprintf(conn,
                  "20%02u-%02u-%02uT%02u:%02u:%02uZ,%f,%f,%u,%s,%s,%s\n",
                  parse_int( dtstr + 4, 2),
                  parse_int( dtstr + 2, 2),
                  parse_int( dtstr, 2),
                  parse_int( timestr, 2),
                  parse_int( timestr + 2, 2),
                  parse_int( timestr + 4, 2),
                  lat,
                  lon,
                  (int)spd,
                  data_buffers[12],
                  data_buffers[32],
                  data_buffers[16]);

    } else {
        eventprintf(conn,
                    "20%02u-%02u-%02uT%02u:%02u:%02uZ,%f,%f,%u,%s\n",
                    parse_int( dtstr + 4, 2),
                    parse_int( dtstr + 2, 2),
                    parse_int( dtstr, 2),
                    parse_int( timestr, 2),
                    parse_int( timestr + 2, 2),
                    parse_int( timestr + 4, 2),
                    lat,
                    lon,
                    (int)spd,
                    data_buffers[34]);
    }
}

void process(connection * conn) {
    conn->current_packet_valid = false;

    //if we've at least got a header
    if (conn->read_count > 5) {
        rep(conn->recv_buffer, 0, ' ', conn->read_count); //remove all null characters up to read count
        size_t index = idx(conn->recv_buffer, '!');

        if (index > 0 && index < conn->read_count) {
            index++;
            process_message(conn, conn->recv_buffer, index);
            memmove(conn->recv_buffer, conn->recv_buffer + index, conn->read_count - index);
            conn->read_count -= index;
            return;
        }
    }

    process_command_file(conn);

    //if we're idle, and 5 minutes have passed we should get status from our device
    if (conn->read_count == 0 && ( time(0) - conn->since_last_status ) > 300) {
        conn->since_last_status = time(0);
    }
}
